﻿using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using TSLab.Script.Helpers;

// ReSharper disable InconsistentNaming
namespace TSLab.Script.Handlers
{
    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class TEMA : DoubleStreamAndValuesHandlerWithPeriod
    {
        private EMA m_ema1;
        private EMA m_ema2;
        private EMA m_ema3;

        public override bool IsGapTolerant
        {
            get { return IsSimple; }
        }

        public override IList<double> Execute(IList<double> source)
        {
            return Calc(source, Period);
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_ema1 = new EMA { Context = Context, Period = Period };
            m_ema2 = new EMA { Context = Context, Period = Period };
            m_ema3 = new EMA { Context = Context, Period = Period };
        }

        protected override void ClearExecuteContext()
        {
            m_ema1 = null;
            m_ema2 = null;
            m_ema3 = null;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - GapTolerancePeriodMultiplier * Period);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                var ema1 = m_ema1.Execute(source, i);
                var ema2 = m_ema2.Execute(ema1, i);
                var ema3 = m_ema3.Execute(ema2, i);
            }
        }

        protected override double Execute()
        {
            if (IsSimple)
                return m_executeContext.Source;

            var ema1 = m_ema1.Execute(m_executeContext.Source, m_executeContext.Index);
            var ema2 = m_ema2.Execute(ema1, m_executeContext.Index);
            var ema3 = m_ema3.Execute(ema2, m_executeContext.Index);
            var result = 3 * ema1 - 3 * ema2 + ema3;
            return result;
        }

        public static IList<double> Calc(IList<double> source, int period)
        {
            if (source == null)
                throw new ArgumentNullException(nameof(source));

            if (period <= 0)
                throw new ArgumentOutOfRangeException(nameof(period));

            var result = new double[source.Count];
            if (result.Length > 0)
            {
                if (period == 1 || result.Length == 1)
                    source.CopyTo(result, 0);
                else
                {
                    var ema1 = Series.EMA(source, period);
                    var ema2 = Series.EMA(ema1, period);
                    var ema3 = Series.EMA(ema2, period);

                    for (var i = 0; i < result.Length; i++)
                        result[i] = 3 * ema1[i] - 3 * ema2[i] + ema3[i];
                }
            }
            return result;
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class DEMA : DoubleStreamAndValuesHandlerWithPeriod
    {
        private EMA m_ema1;
        private EMA m_ema2;

        public override bool IsGapTolerant
        {
            get { return IsSimple; }
        }

        public override IList<double> Execute(IList<double> source)
        {
            if (source == null)
                throw new ArgumentNullException(nameof(source));

            var result = new double[source.Count];
            if (result.Length > 0)
            {
                if (Period == 1 || result.Length == 1)
                    source.CopyTo(result, 0);
                else
                {
                    var ema1 = Series.EMA(source, Period);
                    var ema2 = Series.EMA(ema1, Period);

                    for (var i = 0; i < result.Length; i++)
                        result[i] = 2 * ema1[i] - ema2[i];
                }
            }
            return result;
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_ema1 = new EMA { Context = Context, Period = Period };
            m_ema2 = new EMA { Context = Context, Period = Period };
        }

        protected override void ClearExecuteContext()
        {
            m_ema1 = null;
            m_ema2 = null;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - GapTolerancePeriodMultiplier * Period);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                var ema1 = m_ema1.Execute(source, i);
                var ema2 = m_ema2.Execute(ema1, i);
            }
        }

        protected override double Execute()
        {
            if (IsSimple)
                return m_executeContext.Source;

            var ema1 = m_ema1.Execute(m_executeContext.Source, m_executeContext.Index);
            var ema2 = m_ema2.Execute(ema1, m_executeContext.Index);
            var result = 2 * ema1 - ema2;
            return result;
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    //[HandlerName("TEMA (Zero Lag)")]
    //[HandlerName("TEMA (без задержки)", Language = "ru-RU")]
    [SuppressMessage("StyleCopPlus.StyleCopPlusRules", "SP0100:AdvancedNamingRules", Justification = "Reviewed. Suppression is OK here.")]
    public sealed class ZeroLagTEMA : DoubleStreamAndValuesHandlerWithPeriod
    {
        private TEMA m_tema1;
        private TEMA m_tema2;

        public override bool IsGapTolerant
        {
            get { return IsSimple; }
        }

        public override IList<double> Execute(IList<double> source)
        {
            return Calc(source, Period);
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_tema1 = new TEMA { Context = Context, Period = Period };
            m_tema2 = new TEMA { Context = Context, Period = Period };
        }

        protected override void ClearExecuteContext()
        {
            m_tema1 = null;
            m_tema2 = null;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - GapTolerancePeriodMultiplier * Period);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                var tema1 = m_tema1.Execute(source, i);
                var tema2 = m_tema2.Execute(tema1, i);
            }
        }

        protected override double Execute()
        {
            if (IsSimple)
                return m_executeContext.Source;

            var tema1 = m_tema1.Execute(m_executeContext.Source, m_executeContext.Index);
            var tema2 = m_tema2.Execute(tema1, m_executeContext.Index);
            var result = 2 * tema1 - tema2;
            return result;
        }

        public static IList<double> Calc(IList<double> source, int period)
        {
            if (source == null)
                throw new ArgumentNullException(nameof(source));

            if (period <= 0)
                throw new ArgumentOutOfRangeException(nameof(period));

            var result = new double[source.Count];
            if (result.Length > 0)
            {
                if (period == 1 || result.Length == 1)
                    source.CopyTo(result, 0);
                else
                {
                    var tema1 = TEMA.Calc(source, period);
                    var tema2 = TEMA.Calc(tema1, period);

                    for (var i = 0; i < source.Count; i++)
                        result[i] = 2 * tema1[i] - tema2[i];
                }
            }
            return result;
        }
    }
}
